import subprocess
import os
from typing import List
import toml
import rich.progress as prog
from rich.progress import Progress

def get_zcat_command() -> str:
    '''
    Checks if zcat or gzcat exist on the machine, returns whichever is functional!
    '''
    try:
        subprocess.check_output("zcat --help", shell=True)
    except subprocess.CalledProcessError:
        try:
            subprocess.check_output("gzcat --help", shell=True)
        except subprocess.CalledProcessError:
            raise Exception("Error working with fastq.gz file! zcat or gzcat not found on this machine!")
        else:
            return "gzcat"
    else:
        return "zcat"
    

def check_jellyfish():
    '''
    Checks if jellyfish is installed. Raises an exception if not.
    '''
    try:
        subprocess.check_output("jellyfish --help", shell=True)
    except subprocess.CalledProcessError:
        raise Exception("Jellyfish not found! Please install Jellyfish to count kmers!")


def try_command(command: str, err_msg: str):
    '''
    Runs and command and if there's an error, raises an exception with the provided error message.
    '''
    try:
        subprocess.run(command, shell=True, check=True)
    except subprocess.CalledProcessError as e:
        raise Exception(f"{err_msg}{e}")
    

def handle_config_jellyfish(k: int, file_paths: List[str]) -> dict:
    '''
    Reads a TOML file and runs the commands specified in the 'jellyfish_args' section.
    
    Returns a dictionary of the jellyfish arguments from the config file if parsing the jellyfish
    part of config file succeeds.

    Otherwise, returns None.
    '''

    config_file_path = "config.toml"
    config_keys_valid_options = ["-kmer_length", "-hash_size", "-threads", "-input_file_paths", "-output_file_path"]

    if not os.path.isfile(config_file_path):  
        return None
    
    # read the TOML file
    try:
        config = toml.load(config_file_path)
    except Exception as e:
        if type(e) == toml.TomlDecodeError:
            print("ERROR: Invalid TOML file. Please provide a valid TOML file.")
        exit(0)

    args = {}
    # check if the 'jellyfish_args' section exists
    if 'jellyfish_args' in config:
        jellyfish_args = config['jellyfish_args']
        
        # iterate through the jellyfish_args section of the config file
        for key, value in jellyfish_args.items():

            if key not in config_keys_valid_options:
                # ignore invalid keys but let the user know there was one in the file
                print(f"ERROR: Invalid key: '{key}' was found in the 'jellyfish_args' section of the config file.")
                exit(0)
            elif value != "" and value != None:
                # key and value are valid, add to args dictionary
                # TODO: possibly add validations for type of value it should be
                args[key] = value
        
        # if the required args are not present in the config file and valid, exit the program
        if ('-kmer_length' not in args) or ('-input_file_paths' not in args):

            if ('-kmer_length' not in args) and (k != None) and (k != "") and (k > 0):
                args['-kmer_length'] = k
            if ('-input_file_paths' not in args) and file_paths != None and file_paths != "":
                args['-input_file_paths'] = file_paths
            # k and file_paths are unacceptable values, and the required args are not present in the config file
            if ('-kmer_length' not in args) or ('-input_file_paths' not in args):
                print("ERROR: Kmer length and input file paths are required. Please provide a value for these keys.")
                exit(0)
        
        if (args['-kmer_length'] < 1):
            print("ERROR: Kmer length must be a positive integer.")
            exit(0)

        if (args['-input_file_paths'] == "") or (args['-input_file_paths'] == None):
            print("ERROR: Must provide at least one input file path.")
            exit(0)   
    else:
        # no 'jellyfish_args' section found in the config file
        return None
    
    return args


def fastq_to_kmer_counts(file_paths: List[str],
                         k: int,
                         output_dir: str = "",
                         threads: int = 10, 
                         hash_size: int = 100_000_000) -> str:
    '''
    Takes a path to a 'fastq' or zipped `fastq.gz' file and uses Jellyfish
    to count the provided number of kmers of length 'k'.

    Returns the local path to the output counts file.
    '''
    check_jellyfish()

    # Note that the config file will override the the arguments passed to this funcation
    parse_config_result = handle_config_jellyfish(k, file_paths)

    if parse_config_result != None:
        # use the values from the config file for the 
        k = parse_config_result['-kmer_length']
        file_paths = parse_config_result['-input_file_paths']

        # check if the other values are present in the config file, if not use default values
        if parse_config_result.get('-threads') != None:
            threads = parse_config_result['-threads']
        else:
            print("Threads not found in config file. Using default value of 10.")

        if parse_config_result.get("-hash_size") != None:
            hash_size = parse_config_result['-hash_size']
        else:
            print("Hash size not found in config file. Using default value of 100,000,000.")

        if parse_config_result.get("-output_file_path") != None:
            output_dir = parse_config_result['-output_file_path']
        else:
            print("Output file path not found in config file. Using default value of current directory.")

    # Use the accession number as the base name
    base_path = f"{output_dir}/{os.path.basename(file_paths[0].replace('_1', ''))}"

    # Modify the file extension to .jf for the output
    jf_file = base_path.replace('.fastq', f'_{k}.jf').replace('.gz', '')

    # The base command for kmer counting
    count_command = f"jellyfish count -m {k} -s {hash_size} -C -t {threads} -o {jf_file}"

    # modifies the base command depending on if the files are zipped or not
    if file_paths[0].endswith('.fastq.gz'):
        zcat_command = get_zcat_command()
        count_command = f"{zcat_command} {' '.join(file_paths)} | {count_command} /dev/fd/0"
    else:
        count_command = f"{count_command} {' '.join(file_paths)}"

    # Run count and dump jellyfish commands
    with create_progress_bar() as progress:
        task = progress.add_task(f"Counting {k}-mers...", total=None)

        try_command(count_command, err_msg="Error running Jellyfish count: ")

        counts_file = jf_file.replace(".jf", ".counts")
        dump_command = f"jellyfish dump -c {jf_file} > {counts_file}"

        try_command(dump_command, err_msg="Error running Jellyfish dump: ")
        progress.update(task, total=1, advance=1)
    
    return counts_file

def create_progress_bar() -> Progress:
    '''
    Returns a progress bar configured for k-mer counting progress.
    '''
    return Progress(prog.SpinnerColumn(),
                    prog.TextColumn("[progress.description]{task.description}"),
                    prog.BarColumn(),
                    prog.TimeElapsedColumn())


import typer
import os

import pylagg.cgr as cgr_g
import pylagg.ena_download as ena_d
import pylagg.kmer_counting as kmer_c

app = typer.Typer()

def accession_to_cgr(accession: str, k: int, output_dir: str, threads: int):
    '''
    Takes an accession number and k-mer count and returns a CGR image
    '''
    files = ena_d.ena_download(accession, output_dir) 
    counts_path = kmer_c.fastq_to_kmer_counts(files, k, output_dir=output_dir, threads=threads)

    with open(counts_path, 'r') as f:
        cgr_g.count_file_to_image_file(f, counts_path.replace(".counts", ".png"))

@app.command()
def cgr(
    input: str = typer.Option(
        None,
        "--input",
        "-i",
        help = "File name if using k-mer input you already have. Must be a .txt file for a single image.",
    ),
    accession_number: str = typer.Option(
        None,
        "--accession-number",
        "-a",
        help = 'Generate an image using an NCBI database accession number. If you would like to use more than one accession number, please list them in quotation marks, separated by commas, or put them in a .txt file with one accession number per line, and input the file name with this flag.',
    ),
    kmer: int = typer.Option(
        10,
        "--kmer",
        "-k",
        help = "Specify your desired k-mer number (Only used when generating from an accession number, if your input is already in k-mer form it will be detected)."
    ),
    output_path: str = typer.Option(
        os.getcwd(),
        "--output-path",
        "-o",
        help="Use this to specify an alternate save location for your generated images. If nothing is specified, the default location is where the terminal is located.",
    ),
    scale_size: int = typer.Option(
        100,
        "--scale-size",
        "-s",
        help = "Scale your image down by a percentage as a number between 1 and 100. Do not include any symbols (%)!",
    ),
    thread_count: int = typer.Option(
        16,
        "--thread-count",
        "-t",
        help = "Amount of threads you wish to use. Threads are only used for k-mer counting when generating from an accession number.",
    ),
    config: str = typer.Option(
        None,
        "--config",
        "-c",
        help = "Use a config file to specify your options. Please include only an input (accession number(s) or kmer file(s)) and the config file's name. If any other options are also specified they will be ignored."
        ),
):
    """
    Generate your graph. Type "lagg cgr --help" to see all options for this command.
    """
    
    # INPUT ERROR CHECKING
    if input and accession_number:
        print("Please only include a file name OR an accession number(s).")
        print("If you need help, type 'lagg --help'.")
        exit()
    if not (input or accession_number):
        print("Please include an input, either an accession number(s) or the name of a file containing k-mer input(s).")
        print("If you need help, type 'lagg --help'.")
        exit()
    if scale_size > 100 or scale_size < 1:
        print("Please use a number between 1 and 100 for scale size.")
        print("If you need help, type 'lagg --help'.")
        exit()
    if kmer <= 0:
        print("Invalid k-mer count. Please use a k-mer count greater than 0.")
        print("If you need help, type 'lagg --help'.")
        exit()
    if not(os.path.exists(output_path)):
        print("The given output path does not exist. Please double check your path and try again.")
        print("If you need help, type 'lagg --help'.")
        exit()
    if input and not (os.path.exists(output_path +"/"+ input)):
        print("The input file name is invalid or does not exist. Please make sure your file name includes either '.txt'")
        print("If you need help, type 'lagg --help'.")
        exit()
    if input and not((input.rfind(".txt") != -1)):
        print("Your input is not a supported file type. Supported types are '.txt'. Please convert to a supported file type and try again.")
        print("If you need help, type 'lagg --help'.")
        exit()
        
    # END INPUT ERROR CHECKING
    
    # Pre Process Accession number
    if accession_number:
        
        #remove white spaces
        accession_number = accession_number.replace(" ", "")
        
        accession_list = accession_number.split(",")
            
        #print(accession_list)
        
        if(accession_number.rfind(".txt") != -1):
            print("detecting .txt of accession numbers (not implemented)")
            exit()
        
        # Image generation with accession number:
        #print("Image generation with accession number not yet implemented")
        for number in accession_list:
            accession_to_cgr(number, kmer, output_path, thread_count)
        
        if scale_size != 100:
            print("scaling not yet implemented")
            
    if input:
        counts_file_path = output_path + "/" + input
        with open(counts_file_path) as f:
            
            # .txt case
            if(input.rfind(".txt") != -1):
                if scale_size != 100:
                    print("scaling single kmer input not yet implemented")
                else:
                    # Scale is 100% so generate normal full sized image
                    input = input.replace(".txt", "")
                    #print(input)
                    cgr_g.count_file_to_image_file(f, output_path + "/" + input + ".png")
                    print("\nSuccessfully created image called '" + input + ".png' at " + output_path)
            else:
                #this should never be hit since it's checked above...but just in case
                print("Error with file type")
 
@app.command()               
def ena(
    accession_number: str = typer.Option(
        None,
        "--accession-number",
        "-a",
        help = 'Download a fastq file using an NCBI database accession number. If you would like to use more than one accession number, please list them in quotation marks, separated by commas, or put them in a .txt file with one accession number per line, and input the file name with this flag.',
    ),
    kmer: bool = typer.Option(
        False,
        "--kmer",
        "-k",
        help = "If you would also like your download to be k-mer counted. By default, will not k-mer count."
    ),
    kmer_number: int = typer.Option(
        10,
        "--kmer-number",
        "-n",
        help = "Specify your desired k-mer number, if you want k-mer counting."
    ),
    output_path: str = typer.Option(
        os.getcwd(),
        "--output-path",
        "-o",
        help="Use this to specify an alternate save location for your downloaded files (and k-mer count files, if generating). If nothing is specified, the default location is where the terminal is located.",
    ),
    thread_count: int = typer.Option(
        16,
        "--thread-count",
        "-t",
        help = "Amount of threads you wish to use. Threads are only used for k-mer counting.",
    ),
):
    """
    Only download a fastq file from ENA without generating a graph, and optionally k-mer count it too!
    """
    
    accession_number = accession_number.replace(" ", "")
        
    accession_list = accession_number.split(",")
            
        #print(accession_list)
        
    if(accession_number.rfind(".txt") != -1):
        print("detecting .txt of accession numbers (not implemented)")
        exit()
    
    for number in accession_list:  
        files = ena_d.ena_download(number, output_path)
    
        if kmer:
            counts_path = kmer_c.fastq_to_kmer_counts(files, kmer_number, output_path, thread_count)
            print("Successfully downloaded and created k-mer counts at " + output_path)
        else:
            print("Successfully downloaded file(s) at " + output_path)
            
def cli():
    app()



from ftplib import FTP
import os
from typing import List, Optional
from io import BufferedWriter

import rich.progress as prog
from rich.progress import Progress
import toml

def quit_connection(message: Optional[str], ftp: FTP):
    if message is not None:
        print(message)
    # quit() can throw an exception if ftp server responds with error
    try:
        ftp.quit()
    except Exception as e:
        # This is more like a log message, just something to know that the ftp QUIT command failed.
        # This is not a critical error, so we can just close the connection with ftp.close().
        print("ftp QUIT command failed, trying to close the connection with ftp.close()", e)
        ftp.close()

# def download_with_retries(ftp, file_name, local_file_path, retries, delay):
#     for attempt in range(retries):
#         try:
#             with open(local_file_path, 'wb') as f:
#                 ftp.retrbinary(f"RETR {file_name}", f.write)
#             return True 
#         except ftplib.error_temp as e:
#             print(f"Temporary error: {e}, retrying... ({attempt + 1}/{retries})")
#             # halts execution of the program for a specified number of seconds
#             time.sleep(delay)
#         except Exception as e:
#             return False
#     # if all retries fail, return False
#     return False

def handle_config_ena_download() -> bool:
    '''    
    Reads a TOML file and runs the commands specified in the 'download_args' section.
    
    Returns True if parsing the download part of config file succeeds and results in data, and False otherwise.
    '''

    config_file_path = "config.toml"

    if not os.path.isfile(config_file_path):  
        return False
    
    # read the TOML file
    try:
        config = toml.load(config_file_path)
    except Exception as e:
        if type(e) == toml.TomlDecodeError:
            print("ERROR: Invalid TOML file. Please provide a valid TOML file.")
        return False

    accession_list_config = []
    output_directory_config = None
    if 'download_args' in config:
        arguments = config['download_args']
        
        for key, value in arguments.items():
            if (key == '-list_of_accessions'):
                if isinstance(value, list):
                    accession_list_config = value
                else:
                    print("ERROR: Value for key '-list_of_accessions' in config file must be a list.")
                    return False
            elif (key == '-output_dir'):
                # TODO: possibly add validations for type of value that output_dir should be
                output_directory_config = value
            else:
                print("ERROR: Found unexpected key in download_args section of config file: {}".format(key))
                return False
            
        if (accession_list_config == []):
            print("ERROR: No accession numbers found in the config file.")
            return False
        
        if (output_directory_config != None):
            for item in accession_list_config:
                ena_download(item, output_directory_config)

        if (output_directory_config == None):
            for item in accession_list_config:
                ena_download(item, None)

        return True        

    else:
        # no 'download_args' section found in the config file
        return False


def ena_download(sra_accession: str, output_dir: str = None) -> List[str]: 

    # TODO: what to do in the case that someone provides an accession number in the cli and also
    # provides a config file with a list of accession numbers? should we just apppend that number 
    # to the list if it is not already in the list?

    # small argument validations for the sra_accession parameter
    if (not sra_accession.isalnum()):
        print("Invalid SRA accession number. Please provide a valid SRA accession number.")
        return

    ftp = FTP('ftp.sra.ebi.ac.uk')
    ftp.login()

    prefix = sra_accession[:6]
    last_digit_of_accession = sra_accession[len(sra_accession)-1]

    # handles different format of directory for shorter accession numbers
    if (len(sra_accession) < 10):
        directory = f'/vol1/fastq/{prefix}/{sra_accession}'
    else:
        directory = f'/vol1/fastq/{prefix}/00{last_digit_of_accession}/{sra_accession}'

    try:
        ftp.cwd(directory)
    except Exception:
        quit_connection("Failed to access the directory for the provided accession number.\n"
                 "Please ensure that the accession number is correct and the corresponding\n"
                 "FASTQ files are available on ENA.", ftp)
        return

    file_names = ftp.nlst()
    if (file_names == []):
        quit_connection("No files found for the given SRA accession number.", ftp)
        return
    
    if (output_dir != None):
        if not os.path.exists(output_dir):
            quit_connection("Output directory given for ENA downloading results does not exist.", ftp)
            return

    output_files = []

    with create_progress_bar() as progress:
        for file_name in file_names:
            size = ftp.size(f"{file_name}")
            task = progress.add_task(f"Downloading {file_name}", total=size)
            
            # build local file path
            if (output_dir != None):
                local_file_path = os.path.join(output_dir, file_name)
            else:
                local_file_path = file_name

            output_files.append(local_file_path)
            
            # skip download if the entire file already exists
            if os.path.isfile(local_file_path) and os.path.getsize(local_file_path) == size:
                progress.update(task, advance=size)
                continue

            # Use for later if we want to implement downloads with retries 

            #       if (download_with_retries(ftp, file_name, local_file_path, 3, 2)):
            #           print(f"Downloaded {file_name} successfully!")
            #       else:
            #           print(f"Failed to download {file_name}.")

            with open(local_file_path, 'wb') as f:
                callback = lambda data : write_file(f, data, progress, task)
                ftp.retrbinary(f"RETR {file_name}", callback)

            
            
    quit_connection(None, ftp)
    return output_files


def write_file(file: BufferedWriter, data: bytes, progress: Progress, task: prog.TaskID):
    '''
    Writes data to a file buffer and updates the task for a progress bar
    '''
    file.write(data)
    progress.update(task, advance=len(data))


def create_progress_bar() -> Progress:
    '''
    Creates a progress bar configured for ena downloading.
    '''
    return Progress(prog.SpinnerColumn(),
                    prog.TextColumn("[progress.description]{task.description}"),
                    prog.BarColumn(),
                    prog.DownloadColumn(),
                    prog.TaskProgressColumn(),
                    prog.TimeElapsedColumn())



import csv
import requests


def bulk(taxon_id:str):

    url = f'https://www.ebi.ac.uk/ena/portal/api/links/taxon?accession={taxon_id}&format=tsv&result=read_run'
    response = requests.get(url)

    # Save the content to a file
    if response.status_code == 200:
        with open('ena_taxon_study.tsv', 'wb') as file:
            file.write(response.content)
        print("File downloaded successfully.")
    else:
        print(f"Failed to download the file. Status code: {response.status_code}")

    file_path ='ena_taxon_study.tsv'
    with open(file_path, mode='r', newline='') as file:
        tsv_reader = csv.reader(file, delimiter='\t')
        
        # Skip the header (first line)
        next(tsv_reader)
        
        # Extract the first word of each line
        for row in tsv_reader:
            if row:  # Make sure the row is not empty
                print(row[0]) #print for now but in the future this will be where the function to download fastq files will be



from cutadapt.cli import main
import os
import toml

def trim(input_file: str,config: dict):

    

    command = []  # Initialize an empty list to hold the commands

     
    arguments = config['cutadapt']
    
    # Iterate through each key and its associated list
    for key, value in arguments.items():
        if isinstance(value, list):
            for item in value:
                command.append(f"{key}")

                # Only append the item if it's not an empty string
                if item:
                    command.append(f"{item}")
        else:
            # Append the key
            command.append(f"{key}")

            # Check if value is not an empty list or empty string and append it
            if value not in ["", []]:
                command.append(f"{value}")
    
    if '-o' not in command:
        # Automatically generate the output file name by adding "_trimmed" before the extension
       
        name, ext = os.path.splitext(input_file)
        name, second_ext = os.path.splitext(name)
        output_file= f"{name}_trimmed{second_ext}{ext}"
        command.append('-o')
        out_dir = config.get('Files', {}).get('output',None)


        if out_dir != None:
            output_file =os.path.join(out_dir,output_file)
            command.append(output_file)
        else:
            command.append(output_file)

 
    command.append(input_file)

    print(f"Running command: {' '.join(command)}\n")
    main(command)

    return output_file
    


import os
import toml

from pylagg.cut_adapt import trim
import pylagg.cgr as cgr 
import pylagg.ena_download as ena


def config_accession_to_cgr(accession: str, config_file: str):
  current_dir = os.getcwd()

  if not os.path.isfile(config_file):
        print(f"The file '{config_file}' does not exist.")
        exit(0)

  try:
        config = toml.load(config_file)
  except:
        print("Error in the config file. Please check your formatting and spelling and try again. ")
        exit(0)
  

  output_dir = config.get('Files', {}).get('output',None)


  if output_dir == None:
      output_dir=current_dir
      

  SRA_file = ena.ena_download(accession, output_dir) #needs to be a string not a list

  # Check if the 'cutadapt' section exists
  if 'cutadapt' in config and bool(config['cutadapt']):
    trimmed_file = trim(SRA_file,config)
    if config.get('Files', {}).get('fastq', False) == False:
       os.remove(SRA_file)
    #run jellyfish with trimmed file
    if config.get('Files', {}).get('trim', False) == False:
      os.remove(trimmed_file)

  else:
      print("jellyfish")
      #run Jellyfish with SRA file
  
  #count_path will be the jellyfish produced kmer count
  counts_path = "test.counts"
  with open(counts_path, 'r') as f:
    cgr.config_count_file_to_image_file(f, counts_path.replace(".counts", ".png"),config)
  if config.get('Files', {}).get('counts', False) == False:
    os.remove(counts_path)

  
  



from typing import Dict, Tuple
from io import BytesIO, TextIOWrapper
import math
import os

import numpy as np
import PIL.Image as im
from rich.progress import track

import toml


def contains_valid_characters(input_string: str) -> bool:
    """
    Checks if the given input string contains valid base pair characters
    """    
    for char in input_string:
        if char not in {'A', 'T', 'C', 'G'}:
            return False
    
    return True


def parse_count_file_line(line: str, k: int) -> Tuple[str, int]:
    """
    Reads a line from an input counts file with a given kmer length 'k', returns the tuple: (kmer, count)
    """

    line_split = line.split()
    kmer = line_split[0]
    
    if k > 0 and len(kmer) != k:
        raise Exception(f"The k-mer {kmer} does not match the reported length k={k}.")

    if not contains_valid_characters(kmer):
        raise Exception(f"Invalid k-mer character in k-mer {kmer} (valid characters are A, T, C, G)")

    try:
        count = line_split[1]

        if count == '' or count.isspace():
            raise Exception()
    except:
        raise Exception(f"Count not provided for k-mer: '{kmer}'")

    try:
        count = int(count)
    except ValueError:
        raise Exception(f"Count for k-mer {kmer} must be an integer'")

    if count < 1:
        raise Exception("All k-mer counts must be ≥1")
    
    return (kmer, count)


def count_file_to_dictionary(file: TextIOWrapper) -> Dict[str, int]:
    """
    Takes a counts file as input and outputs a dictionary representation used later for image generation
    """

    k_dict = {}

    with file:
        k = 0

        file.seek(0)

        for line in file:
            if line.isspace() or line == '':
                continue

            (kmer, count) = parse_count_file_line(line, k)
            k_dict.update({kmer : count})

            if k == 0:
                k = len(kmer)

    return k_dict


def calculate_pos(kmer: str, size: int, corner_labels: Tuple[str, str, str, str]) -> Tuple[int, int]:
    """
    Returns the pixel position (x, y) of a kmer in a CGR image with a given size
    """

    x, y = 0, 0

    # use bit shifting instead of division to avoid floating point values
    offset = size >> 1

    bot_left, _, top_right, bot_right = corner_labels

    for base in reversed(kmer):
        if base == top_right or base == bot_right:
            x += offset

        if base == bot_left or base == bot_right:
            y += offset

        offset >>= 1

    return (x, y)


def generate_image_arr(k_dict: Dict[str, int], verbose=True, size=None, log10=True, normalized=True) -> np.ndarray:
    """
    Generates a numpy array representing an image covering RGB channels
    """
    
    k = len(next(iter(k_dict)))  # gets the length of the first key in k_dict

    if size is None:
        size = 2 ** k

    r = np.zeros((size, size))
    g = np.zeros((size, size))
    b = np.zeros((size, size))

    for kmer, count in track(k_dict.items(), disable=not verbose, description="Generating image..."):
        if log10:
            count = math.log10(count)

        # weak H-bonds W = {A, T} and strong H-bonds S = {G, C} on the diagonals
        r_pos = calculate_pos(kmer, size, ('A', 'G', 'T', 'C'))
        r[r_pos] = count
        
        # purine R = {A, G} and pyrimidine Y = {C, T} on the diagonals
        g_pos = calculate_pos(kmer, size, ('A', 'T', 'G', 'C'))
        g[g_pos] = count

        # amino group M = {A, C} and keto group K = {G, T} on the diagonals
        b_pos = calculate_pos(kmer, size, ('A', 'T', 'C', 'G'))
        b[b_pos] = count

    rgb = np.dstack((r, g, b))

    if normalized:
        rgb = rgb / np.max(rgb)

    return (rgb * 255).astype(np.uint8)


def count_file_to_image(input_data: TextIOWrapper, verbose=True, **kwargs) -> im.Image:
    """
    Takes a counts file as input and returns the generated image as an image object
    """

    k_dict = count_file_to_dictionary(input_data)
    chaos = generate_image_arr(k_dict, verbose, **kwargs)
    return im.fromarray(chaos)


def count_file_to_image_file(input_data: TextIOWrapper, output_file: str | BytesIO, output_type="png", verbose=True):
    """
    Takes counts file data and creates an image at the provided file path or buffer with the given output file type
    """
    
    img = count_file_to_image(input_data, verbose)
    img.save(output_file, output_type)


def config_count_file_to_image_file(input_data: TextIOWrapper, output_file: str | BytesIO, config:dict, output_type="png", verbose=True):
    """
    Takes counts file data and creates an image at the provided file path or buffer with the given output file type
    """
    
    config_size = config.get('cgr', {}).get('size', None)  
    config_log = config.get('cgr', {}).get('log10', True)
    config_normalized = config.get('cgr', {}).get('normalized', True)
    config_output = config.get('Files', {}).get('output', None)
    if config_output != None:
        config_output = os.path.join(config_output, output_file)
        output_file = config_output
    
    img = count_file_to_image(input_data, verbose,size=config_size,normalized = config_normalized)
    img.save(output_file, output_type)

    